#
# Copyright 2009 FUJITSU LIMITED
# All Rights Reserved.
#
# NAME
#      vvcheck.awk - check VersionView.sav for consistency
#
# SYNOPSIS
#      gawk [variables] -f vvcheck.awk file
#
# DESCRIPTION
#      This gawk(1) script checks a VersionView.sav for consistency
#      It examines the structure of the file and its entries for
#      keyword uniqueness, correctness, and completeness.
#      Errors are logged to stdout with the respective line as context.
#      It finally also prints variable statistics.
#
#    Variables
#      Variables may be passed as -v "variable=value" in any order.
#
#      FILEFORMAT={W|L}
#        Setting FORMAT to 'W' (default is 'L') forces file format
#        interpretation according to Windows conventions.
#
#    Input
#      Expected input is formatted according to VersionView.sav(5).
#      The first non-empty line must be a valid section.
#
#    Output
#      Problems found in the input are logged to stdout as follows
#      (NR and $n are meant as usual in gawk):
#
#        NR: $0 in section <section>
#            <class>: <descriptive text>
#            ...
#
#      where <class> is one of {"ERROR" | "INFO"}.
#
#    Implementation Assumptions and Considerations
#      1. Input is from ASCII character set.
#      2. This script needs gawk 3.1 extensions:
#         systime(), strftime(), strtonum()
#      3. Script should work VersionView.sav files for Windows and
#         Linux. The file format is guessed from its contents:
#      4. Variables named with 1st capital letter indicate program globals.
#         Variables configurable via command line are all caps.
#
# EXIT STATUS
#      The script exits with status zero after sucessful execution and
#      aborts with status non-zero in case of errors in the input to
#      process
#
# SEE ALSO
#      gawk(1), VersionView.sav(5)
#


# ----------------------------------------------------------------
BEGIN {
# ----------------------------------------------------------------

	FS = "="	# = is significant in this file, not space
	OK = 0
	FAILED = 1

	if (FILEFORMAT == "")
		FILEFORMAT = "L"

	FSC = "FUJITSU"
	VVAgt["L"] = "ServerView SNMP VersionView Agent"
	VVAgt["W"] = "VersionView Agent"

	PreviousNR = 0				# previously logged line NR
	FileFormat = 1				# 1 = Linux 0 = Windows
	Need1stLine = 1				# 1 = TRIE
	ExitStatus = 0				# 0 = OK
	TreeLevel = 2
	Agents = 0
	KeysFound = 5

	# Section Name				| Section Contents
	#
	# [Last Date]				| Info about the state of the file
	# [Content Info]			| Info about the content of the file
	# [Agent Info]				| Info about the version of the agent
	# [Component Definition]	| Definition of identifiers for components
	# [Physical Definition]		| ... physical containers
	# [Functional Definition]	| ... and functional containers
	# [Physical Container]		| The physical containment hierarchy
	# [Functional Container]	| The functional containment hierarchy
	# [Component Table]			| The components found
	# [Firmware Update]			| Components whose FW/SW can be updated

	Sections["Last Date"] = 0
	Sections["Agent Info"] = 0
	Sections["Component Definition"] = 0
	Sections["Physical Definition"] = 0
	Sections["Functional Definition"] = 0
	Sections["Physical Container"] = 0
	Sections["Functional Container"] = 0
	Sections["Component Table"] = 0
	Sections["Firmware Update"] = 0
	Sections["Content Info"] = 0

	CurrSection = ""
	InvalidSections = ""
	HaveContentInfo = 0

	# [Last Date]

	LastDate["LastCollectionDate"] = ""
	LastDate["LastDate"] = ""
	LastDate["LastScanDate"] = ""
	LastDate["LastScan"] = ""
	LastDate["ScanState"] = ""

	# [Agent Info]

	AgentInfo["AgentId"] = ""
	AgentInfo["AgentCompany"] = ""
	AgentInfo["AgentVersion"] = ""
	AgentInfo["AgentBuild"] = ""

	# [Content Info]
	# [0, <keyword>] = -1: <keyword> not in [Content Info]
	# [0, <keyword>] =  n: <keyword>=n exists in [Content Info]
	# [1, <keyword>] =  m: <keyword> instance counter in file

	for (i = 0; i < 2; i++) {
		ContentInfo[i, "CompDefFound"] = i - 1
		ContentInfo[i, "PhysNodeDefFound"] = i - 1
		ContentInfo[i, "PhysFuncDefFound"] = i - 1
		ContentInfo[i, "PhysContDefFound"] = i - 1
		ContentInfo[i, "FuncContDefFound"] = i - 1
		ContentInfo[i, "InfoTableDefFound"] = i - 1
		ContentInfo[i, "InfoFwObjectDefFound"] = i - 1

		ContentInfo[i, "NrPContainer"] = i - 1
		ContentInfo[i, "NrFContainer"] = i - 1
		ContentInfo[i, "NrComponents"] = i - 1
		ContentInfo[i, "NrUpdObjects"] = i - 1
	}

	# [Component Definition]
	# The defined components

	ComponentDefinition["00"] = "No class"
	ComponentDefinition["01"] = "Chassis"
	ComponentDefinition["02"] = "System board"
	ComponentDefinition["03"] = "Main memory module"
	ComponentDefinition["04"] = "Arbiter configuration"
	ComponentDefinition["05"] = "Management controller"
	ComponentDefinition["06"] = "Cabinet"
	ComponentDefinition["07"] = "UPS"
	ComponentDefinition["08"] = "Power supply"
	ComponentDefinition["09"] = "Power supply back plane"
	ComponentDefinition["0A"] = "CPU"
	ComponentDefinition["0B"] = "I/O"
	ComponentDefinition["0C"] = "SCSI Controller"
	ComponentDefinition["0D"] = "LAN / Network controller"
	ComponentDefinition["0E"] = "WAN controller"
	ComponentDefinition["0F"] = "Memory expander"
	ComponentDefinition["10"] = "Graphic controller"
	ComponentDefinition["11"] = "Environment controller"
	ComponentDefinition["12"] = "SCSI hard disk back plane"
	ComponentDefinition["13"] = "Fan"
	ComponentDefinition["14"] = "SCSI Termination Module"
	ComponentDefinition["15"] = "Complete System"
	ComponentDefinition["16"] = "Memory Termination Module"
	ComponentDefinition["17"] = "Front Panel"
	ComponentDefinition["18"] = "LVD-Extender"
	ComponentDefinition["19"] = "SCSI Device"
	ComponentDefinition["1A"] = "Bridge Device"
	ComponentDefinition["1B"] = "Management Controller Board"
	ComponentDefinition["1C"] = "Microcontroller"
	ComponentDefinition["1D"] = "VRM"
	ComponentDefinition["1E"] = "DC_DC Converter"
	ComponentDefinition["1F"] = "Management Blade"
	ComponentDefinition["A0"] = "BIOS"
	ComponentDefinition["A1"] = "Firmware"
	ComponentDefinition["A2"] = "Device driver"
	ComponentDefinition["A3"] = "Management agent"
	ComponentDefinition["A4"] = "Software component"
	ComponentDefinition["A5"] = "Fibrechannel back plane"
	ComponentDefinition["A6"] = "Fibre Link Adapter"
	ComponentDefinition["A7"] = "Operating System"
	ComponentDefinition["A8"] = "ServerControl"
	ComponentDefinition["A9"] = "SDRVersion"
	ComponentDefinition["FF"] = "Reserved"

	ComponentDefinition2[0] = 0

	# [Physical Definition] and [Functional Definition]
	# Type.1=1
	# Name.1=OtherContainer
	# Type.2=2
	# Name.2=PhysPrimergy
	# Type.3=3
	# Name.3=ServerCabinet
	# Type.4=4
	# Name.4=StorageExtension

	PhysicalDefinition[1] = "OtherContainer"
	PhysicalDefinition[2] = "PhysPrimergy"
	PhysicalDefinition[3] = "ServerCabinet"
	PhysicalDefinition[4] = "StorageExtension"

	FunctionalDefinition[1] = "OtherContainer"
	FunctionalDefinition[2] = "FuncPrimergy"
	FunctionalDefinition[3] = "FuncChassis"
	FunctionalDefinition[4] = "SubsystemContainer"
	FunctionalDefinition[5] = "BaseBoard"
	FunctionalDefinition[6] = "SCSISystem"
	FunctionalDefinition[7] = "SCSIDevice"
	FunctionalDefinition[8] = "LANSystem"
	FunctionalDefinition[9] = "WANSystem"
	FunctionalDefinition[10] = "GraphicSystem"
	FunctionalDefinition[11] = "Software"
	FunctionalDefinition[12] = "BladeSystem"	# not in 3.02 or before

	# [Physical Container] and [Functional Container]
	# Arrays to catch their state

	PhysicalContainer[0, "type"] = 0
	PhysicalContainer[0, "name"] = 0
	PhysicalContainer[0, "parent"] = 0
	PhysicalContainer[0, "comps"] = 0
	PhysicalContainer[0, "children"] = 0

	FunctionalContainer[0, "type"] = 0
	FunctionalContainer[0, "name"] = 0
	FunctionalContainer[0, "parent"] = 0
	FunctionalContainer[0, "comps"] = 0
	FunctionalContainer[0, "children"] = 0
}

# ----------------------------------------------------------------
# Main Program
# ----------------------------------------------------------------

/./ && Need1stLine {
	# Test the 1st non-empty line.
	# Refuse this file, if it is not a [<section>].

	if ($0 !~ /\[[a-zA-Z ]+\]/) {
		logThisLine()
		printf("\tERROR: The 1st non-empty line is not a [<section>]. Invalid file. Exit.\n")
		ExitStatus = 1
		exit
	}

	Need1stLine = 0

	# fall thru ...
}

# Identify and remember section heads
# Completeness is tested in END.

/^[[:space:]]*\[[[:alpha:] ]+\][[:space:]]*$/ {

	pos = match($0, "[a-zA-Z ]+")
	CurrSection = substr($0, RSTART, RLENGTH)

	if (CurrSection in Sections) {
		if (Sections[CurrSection] != 0) {
			printf("%d: %s\n", NR, $0)
			printf("\tERROR: Section [%s] is already defined!\n", CurrSection)
		} else
			Sections[CurrSection]++

	} else {
		InvalidSections = InvalidSections "[" CurrSection "] "
		printf("%d: %s\n", NR, $0)
		printf("\tERROR: Section [%s] %s\n", CurrSection, "is undefined!")
	}

	Need1stLine = 0
	next
}

# [Last Date]
# LastCollectionDate=2005-10-17 13:18:18
# LastDate=4353887A
# LastScanDate=2005-10-24 10:45:00
# LastScan=435C9F0C
#
# Test keyword existence and uniqueness and date and second
# values for format plausibility.
# Completeness is tested in END.

$1 == "LastCollectionDate" ||  $1 == "LastScanDate" {

	testTheSection("Last Date")
	testUniqType(LastDate, $1, $2)

	if (testTheDate(1, $2)) {
		logThisLine()
		printf("\tERROR: '%s' is not '20YY-MM-DD hh:mm:ss'!\n", $2)
	}
	next
}

$1 == "LastDate" ||  $1 == "LastScan" {

	testTheSection("Last Date")
	testUniqType(LastDate, $1, $2)

	if (testTheSeconds($2)) {
		logThisLine()
		printf("\tERROR: '%s' has more than 8 or invalid hexdigits!\n", $2)
	} else {
		# NOTE:
		#	This requires gawk(1) strftime() which also evaluates
		#	the system's timezone information (tzset()).

		# logThisLine()
		if ($1 == "LastDate")
			printf("Last file update '%s'\n",
				strftime("%Y-%m-%d-%H:%M:%S", strtonum("0x" $2)))
		if ($1 == "LastScan")
			printf("Last scan '%s'\n",
				strftime("%Y-%m-%d-%H:%M:%S", strtonum("0x" $2)))

	}
	next
}

$1 == "ScanState" {

	testTheSection("Last Date")
	testUniqType(LastDate, $1, $2)

# 	if (FILEFORMAT == "L") {
# 		logThisLine()
# 		printf("\tERROR: keyword '%s' undefined for Linux format\n", $1)
# 	} else
# 		FILEFORMAT = "W"
	next
}

# [Agent Info]
# AgentId=ServerView SNMP VersionView Agent
# AgentCompany=FUJITSU
# AgentVersion=3.20-04  or  AgentVersion=4.11.06.02
# AgentBuild=2005/01/19-12:09
#
# Test keyword existence and uniqueness and values for plausibility.
# Completeness is tested in END.

$1 == "AgentId" || $1 == "AgentCompany" ||
	$1 == "AgentVersion" || $1 == "AgentBuild" {

	testTheSection("Agent Info")
	testUniqType(AgentInfo, $1, $2)

	# Since V4.80, VersionView AgentId for Linux and Windows is identical

	if ($1 == "AgentId" && $2 != VVAgt["L"] && $2 != VVAgt["W"]) {
		logThisLine()
		printf("\tERROR: '%s' is neither '%s' or '%s'!\n", $2, VVAgt["L"], VVAgt["W"])
	}

	if ($1 == "AgentCompany" && $2 != FSC) {
		logThisLine()
		printf("\tERROR: '%s' is not '%s'!\n", $2, FSC)
	}

	if ($1 == "AgentVersion") {
		if (FILEFORMAT == "L" && match($2, "[0-9].[0-9][0-9]-[0-9][0-9]") == 0) {
			logThisLine()
			printf("\tWARNING: '%s' is not 'V.vv-rr'!\n", $2)
		}

		if (FILEFORMAT == "W" && match($2, "[0-9].[0-9][0-9].[0-9][0-9].[0-9][0-9]") == 0) {
			logThisLine()
			printf("\tWARNING: '%s' is not 'V.vv.rr.ss'!\n", $2)
		}
	}

	if ($1 == "AgentBuild") {
		if (FILEFORMAT == "L" && testTheDate(2, $2)) {
			logThisLine()
			printf("\tWARNING: '%s' is not in '20YY/MM/DD-hh:mm'!\n", $2)
		}

		if (FILEFORMAT == "W" && testTheDate(3, $2)) {
			logThisLine()
			printf("\tWARNING: '%s' is not 'Build: MMM DD 20YY, HH:MM:SS'!\n", $2)
		}
	}
	next
}

# [Content Info]
# CompDefFound=43
# PhysNodeDefFound=4
# PhysFuncDefFound=12
# PhysContDefFound=2
# FuncContDefFound=13
# InfoTableDefFound=922
# InfoFwObjectDefFound=7
#
# Test keyword existence and uniqueness. Values are checked in END.
# Completeness is tested in END.

$1 == "CompDefFound" || $1 == "PhysNodeDefFound" || $1 == "PhysFuncDefFound" ||
	$1 == "PhysContDefFound" || $1 == "FuncContDefFound" ||
	$1 == "InfoTableDefFound" || $1 == "InfoFwObjectDefFound" {

	testTheSection("Content Info")
	HaveContentInfo = 1
	ContentInfo[0, $1] = $2
	next
}

# [Component Definition]
# <nn>=<string>
#
# Test keyword existence, uniqueness, and value correctness.
# Completeness is tested in END.

$1 ~ /^[0-9a-fA-F][0-9a-fA-F]$/ {

	testTheSection("Component Definition")
	testUniqType(ComponentDefinition2, $1, $2)

	if ($1 in ComponentDefinition) {
		if ($2 != ComponentDefinition[$1]) {
			logThisLine()
			printf("\tERROR: value '%s' must be '%s'!\n", $2, ComponentDefinition[$1])
		} else
			ContentInfo[1, "CompDefFound"]++

	} else {
		logThisLine()
		printf("\tERROR: keyword '%s' is undefined for [Component definition]!\n", $1)
	}
	next
}

# [Physical Definition] and [Functional Definition]
# Type.<n>=<n>		<n> increasing
# Name.<n>=<value>
#
# [Physical Container] and [Functional Container]
# Type.<n>=<j>		<n> increasing, <j> from within [Physical Definition]
# Parent.<n>=<k>	<k> = 0 or from within this container
# Name.<n>=<value>
#
# Check existence, uniqueness and correctness of both keywords, <n>, <value>.
# Completeness is tested in END. Type should always be first in each group,
# it is the counting instance for the statistical checks.

$1 ~ /^Type\.[0-9]+$/ || $1 ~ /^Name\.[0-9]+$/ {

	testTheSection("Physical Definition|Functional Definition|Physical Container|Functional Container")

	i = split($1, arr, ".")
	n = arr[2]

	if (arr[1] == "Type") {
		if (CurrSection == "Physical Definition") {
			if (n < 1 || n > 4 || n != $2) {
				logThisLine()
				printf("\tERROR: keyword '%s' has '%s' outside 1 - 4 or invalid value '%s'!\n",
					$1, n, $2)
			} else {
				# Remember valid <n> for later reference by creating <array>[n] = "".

				PhysicalDefinition2[n] = ""
				ContentInfo[1, "PhysNodeDefFound"]++
			}

		} else if (CurrSection == "Functional Definition") {
			if (n < 1 || n > 12 || n != $2) {
				logThisLine()
				printf("\tERROR: keyword '%s' has '%s' outside 1 - 12 or invalid value '%s'!\n",
					$1, n, $2)
			} else {
				# Remember valid <n> for later reference by creating <array>[n] = "".

				FunctionalDefinition2[n] = ""
				ContentInfo[1, "PhysFuncDefFound"]++
			}

		} else if (CurrSection == "Physical Container") {
			ContentInfo[1, "PhysContDefFound"]++
			if (testTheN(n, "PhysContDefFound") == OK) {
				PhysicalContainer[n, "type"] = $2
				PhysicalContainer[n, "comps"] = 0
				PhysicalContainer[n, "children"] = 0
			}
			testTheValue($1, $2, PhysicalDefinition, "Physical Definition")

		} else  if (CurrSection == "Functional Container") {
			ContentInfo[1, "FuncContDefFound"]++
			if (testTheN(n, "FuncContDefFound") == OK) {
				FunctionalContainer[n, "type"] = $2
				FunctionalContainer[n, "comps"] = 0
				FunctionalContainer[n, "children"] = 0
				FunctionalContainer[n, "163"] = 0
			}
			testTheValue($1, $2, FunctionalDefinition, "Functional Definition")

		} else {
			printf("%d: %s\n", NR, $0)
			printf("\tERROR: keyword '%s' outside of permitted sections!\n", $1)
		}
	}

	# Now verify in "Name.<n>=<value>" that <n> is as in "Type.<n>"

	if (arr[1] == "Name") {
		if (CurrSection == "Physical Definition") {
			testTheName(n, "PhysNodeDefFound", $2, PhysicalDefinition, PhysicalDefinition2)

		} else if (CurrSection == "Functional Definition") {
			testTheName(n, "PhysFuncDefFound", $2, FunctionalDefinition, FunctionalDefinition2)

		} else if (CurrSection == "Physical Container") {
			testTheN(n, "PhysContDefFound")
			PhysicalContainer[n, "name"] = $2

		} else if (CurrSection == "Functional Container") {
			testTheN(n, "FuncContDefFound")
			FunctionalContainer[n, "name"] = $2

		} else {
			printf("%d: %s\n", NR, $0)
			printf("\tERROR: keyword '%s' outside of permitted sections!\n", $1)
		}
	}
	next
}

# [Physical Container] and [Functional Container]
# Parent.<n>=<k>	<k> = 0 (root) or BACKWARD ref. into [Physical Container]

$1 ~ /^Parent\.[0-9]+$/ {

	testTheSection("Physical Container|Functional Container")

	i = split($1, arr, ".")
	n = arr[2]

	if (CurrSection == "Physical Container") {
		if (testTheN(n, "PhysContDefFound") == OK)
			PhysicalContainer[n, "parent"] = $2

		if ($2 == 0 || testTheType($1, $2, PhysicalContainer, "Physical Container") == OK)
			PhysicalContainer[n, "children"]++


	} else if (CurrSection == "Functional Container") {
		if (testTheN(n, "FuncContDefFound") == OK)
			FunctionalContainer[n, "parent"] = $2

		if ($2 == 0 || testTheType($1, $2, FunctionalContainer, "Functional Container") == OK)
			FunctionalContainer[n, "children"]++

	} else {
		logThisLine()
		printf("\tERROR: keyword '%s' outside of permitted sections!\n", $1)
	}
	next
}

$1 ~ /^NrContainer/ {

	testTheSection("Physical Container|Functional Container")

	# Since (at least) V4.52, there are "NrContainer" also in Linux format
	# if (FILEFORMAT == "L") {
	# 	logThisLine()
	#	printf("\tERROR: keyword '%s' undefined for Linux format\n", $1)
	# } else {
		if (CurrSection == "Physical Container")
			ContentInfo[0, "NrPContainer"] = $2
		if (CurrSection == "Functional Container")
			ContentInfo[0, "NrFContainer"] = $2
	# }
	next
}

# [Component Table]
# PhysCont.<n>=<i>	<i> from within [Physical Container]
# FuncCont.<n>=<j>	<j> from within [Functional Container]
# CompType.<n>=<k>	<k> from within [Component Definition]
# <other>.<n>=<x>
#
# Check that <n> increases. Assume that PhysCont is always first.

$1 ~ /^PhysCont\.[0-9]+$/ || $1 ~ /^FuncCont\.[0-9]+$/ || $1 ~ /^CompType\.[0-9]+$/ {

	testTheSection("Component Table")

	split($1, arr, ".")

	if ($1 ~ "^PhysCont")
		ContentInfo[1, "InfoTableDefFound"]++

	testTheN(arr[2], "InfoTableDefFound")

	if (arr[1] == "PhysCont") {
		if (testTheType($1, $2, PhysicalContainer, "Physical Container") == OK)
			PhysicalContainer[$2, "comps"]++
	}

	if (arr[1] == "FuncCont") {
		if (testTheType($1, $2, FunctionalContainer, "Functional Container") == OK)
			FunctionalContainer[$2, "comps"]++
	}

	if (arr[1] == "CompType") {
		# The indices in ComponentDefinition[] are 2-hexdigit numbers

		testTheValue($1, sprintf("%02X", $2), ComponentDefinition, "Component Definition")
		if ($2 == 163)
			Agents++
	}
	next
}

$1 ~ /^Vendor\.[0-9]+$/ || $1 ~ /^ProdName\.[0-9]+$/ || $1 ~ /^ProdNumber\.[0-9]+$/ ||
	$1 ~ /^ProdDescript\.[0-9]+$/ || $1 ~ /^ManuDate\.[0-9]+$/ || $1 ~ /^Version\.[0-9]+$/ ||
	$1 ~ /^SerialNo\.[0-9]+$/ || $1 ~ /^Language\.[0-9]+$/ || $1 ~ /^CompId\.[0-9]+$/ {

	testTheSection("Component Table")
	next
}

$1 ~ /^NrComponents/ {

	testTheSection("Component Table")

	# Since (at least) V4.52, there are "NrComponents" also in Linux format
	# if (FILEFORMAT == "L") {
	#	logThisLine()
	#	printf("\tERROR: keyword '%s' undefined for Linux format\n", $1)
	# } else {
		ContentInfo[0, "NrComponents"] = $2
	# }
	next
}

# [Firmware Update]
# ObjectName.<n>=<value>
# ObjectParent.<n>=<value>
# ObjectPath.<n>=<value>
# UpdateState.<n>=<value>
# CompIndex.<n>=<k>	<k> from within [Component Definition]
#
# Assume that ObjectName is always first. Check that <n> increases.
# Check that the 5-keyword sets are complete (do not tes the last one).

$1 ~ /^ObjectName\.[0-9]+$/ || $1 ~ /^ObjectParent\.[0-9]+$/ || $1 ~ /^ObjectPath\.[0-9]+$/ ||
	$1 ~ /^UpdateState\.[0-9]+$/ || $1 ~ /^CompIndex\.[0-9]+$/ {

	testTheSection("Firmware Update")

	split($1, arr, ".")

	if ($1 ~ "^ObjectName") {
		ContentInfo[1, "InfoFwObjectDefFound"]++
		if (KeysFound < 5) {
			logThisLine()
			printf("\tERROR: incomplete keyword set before '.%d' in [Firmware Update]\n", arr[2])
		}
		KeysFound = 0
	}
	KeysFound++
	testTheN(arr[2], "InfoFwObjectDefFound")

	# The value of CompIndex.<n> must indicate the number of an entry in the
	# [Component Table]. Just test it against its counter.
	# NOTE: This assumes the [Component Table] was already processed

	if ($1 ~ "^CompIndex") {
		if ($2 > ContentInfo[1, "InfoTableDefFound"]) {
			logThisLine()
			printf("\tERROR: value '%s' undefined in [Component Table]!\n", $2)
		}
	}

	next
}

$1 ~ /^NrUpdObjects/ {

	testTheSection("Firmware Update")

	# Since (at least) V4.52, there are "NrUpdObjects" also in Linux format
	# if (FILEFORMAT == "L") {
	# 	logThisLine()
	# 	printf("\tERROR: keyword '%s' undefined for Linux format\n", $1)
	# } else {
		ContentInfo[0, "NrUpdObjects"] = $2
	# }
	next
}

# Everything else

/./ {
	logThisLine()
	printf("\tERROR: Undefined keyword '%s'!\n", $1)
	next
}

# ----------------------------------------------------------------
# Functions
# ----------------------------------------------------------------

function logThisLine() {
	# Log the current line ONCE to give ERROR context.

	if (PreviousNR < NR) {
		printf("%d: %s in section [%s]\n", NR, $0, CurrSection)
		PreviousNR = NR
	}
}

function logEND() {
	# Log summary statements in the END pattern

	if (PreviousNR != "END") {
		printf("END:\n")
		PreviousNR = "END"
	}
}

function testTheSection(validSections,		idx, sect, array) {
	# We verify that the current section (we are in) is in the valid
	# sections passed as | separated values in the argument

	sect = split (validSections, array, "|")

	for (idx = 1; idx <= sect; idx++) {
		if (CurrSection == array[idx])
			return OK
	}

	logThisLine()
	printf("\tERROR: keyword invalid in section [%s], expected in",
		CurrSection)
	for (sect in array)
		printf(" %s", array[sect])
	printf("!\n")

	return FAILED
}

function testTheDate(format, theDate,	p) {
	# We verify format for format plausibility only according to
	# 	format = 1: 2005-10-17 17:37:01
	# 	format = 2: 2005/01/19-12:09
	#   format = 3: Build: Sep 12 2005, 15:49:38
	# Valid dates are from 2000 thru 2099. Months 00, 13-19,
	# days 00, 32-39, hours 24-29 are accepted w.r.t. simplicity.
	# Note that we allow seconds 0 ... 60 to support leap seconds!

	if (format == 1) {
		p = match(theDate, "20[0-9][0-9]-[01][0-9]-[0-3][0-9] [012][0-9]:[0-5][0-9]:[0-6][0-9]")
		if (p == 0)
		 	return FAILED
	}
	if (format == 2) {
		p = match(theDate, "20[0-9][0-9]/[01][0-9]/[0-3][0-9]-[012][0-9]:[0-5][0-9]")
		if (p == 0)
		 	return FAILED
	}
	if (format == 3) {
		p = match(theDate, "Build: [A-Z][a-z][a-z] [0-3][0-9] 20[0-9][0-9], [012][0-9]:[0-5][0-9]:[0-5][0-9]")
		if (p == 0)
		 	return FAILED
	}
	return OK
}

function testTheSeconds(theSeconds) {
	# We verify for plausibility only: 8 hex digits

	if (length(theSeconds) != 8 || match(theSeconds, "^[[:xdigit:]]+$") == 0)
		return FAILED
	return OK
}


function testTheType(key, value, array, section) {
	# We verify that the type is defined for the given section

	if (!((value, "type") in array)) {
		logThisLine()
		printf("\tERROR: keyword '%s' value '%s' undefined in [%s]!\n",
			$1, value, section)
		return FAILED
	}
	return OK
}

function testTheValue(key, value, array, section) {
	# We verify that the key is defined for the given section

	if (!(value in array)) {
		logThisLine()
		printf("\tERROR: keyword '%s' value '%s' undefined in [%s]!\n",
			key, value, section)
		return FAILED
	}
	return OK
}

function testUniqType(array, idx, value) {
	# For uniqeness, array[idx] must be empty

	if ((idx in array) && array[idx] != "") {
		logThisLine()
		printf("\tERROR: type '%s' is already defined!\n", idx)
		return FAILED
	} else
		array[idx] = value
	return OK
}

function testTheN(theN, counter) {
	# We verify that the <n> is as expected by the corresponding counter

	if (theN != ContentInfo[1, counter]) {
		logThisLine()
		printf("\tERROR: keyword '%s' has '%s' out of order, expected is '%s'!\n",
			$1, theN, ContentInfo[1, counter])
		return FAILED
	}
	return OK
}

function testTheName(theN, counter, value, array, array2) {
	# We verify that <n> in Name.<n> correponds to previous Type.<n>
	# is defined and that <value> in Name.<n>=<value> is correct.

	testTheN(theN, counter)
	testUniqType(array2, theN, value)

	if (theN in array) {
		if (value != array[theN]) {
			logThisLine()
			printf("\tERROR: value '%s' must be '%s'!\n", value, array[theN])
			return FAILED
		}
	} else {
		logThisLine()
		printf("\tERROR: number '%s' is undefined!\n", theN)
		return FAILED
	}
	return OK
}

function compareTheCounter(counter) {
	# we compare count of entries actually found to the count recorded

	if (ContentInfo[0, counter] == -1) {
		logEND()
		printf("\tERROR: Missing in [Content Info]: %s (expected %d)\n",
			counter, ContentInfo[1, counter])
		return FAILED
	}

	if (ContentInfo[1, counter] != ContentInfo[0, counter]) {
		logEND()
		printf("\tERROR: Found %5d [Component Definitions], expected %d\n",
			ContentInfo[1, counter], ContentInfo[0, counter])
		return FAILED
	}
	return OK
}

function printChildren(array, parent,		idx, cnt) {
	# we print the parent/child relationship recursively

	for (idx = 1; ((idx, "parent") in array); idx++) {
		if (array[idx, "parent"] == parent) {
			printf("%s%-24s%s %4d",
				substr("                ", 1, TreeLevel),
				array[idx, "name"],
				substr("                ", 1, 16 - TreeLevel),
				array[idx, "comps"])

			cnt += array[idx, "comps"]
			if (array[idx, "type"] == 11)
				printf(" (%d agents)\n", Agents)
			else
				printf("\n")

			if (array[idx, "children"] > 0) {
				TreeLevel += 2
				cnt += printChildren(array, idx)
				TreeLevel -= 2
			} else
				return cnt
		}
	}
	return cnt
}

# ----------------------------------------------------------------
END {
# ----------------------------------------------------------------

	if (ExitStatus == 1)
		exit ExitStatus

	status = 0

	# [section]
	# Report on missing or duplicate sections.

	missing = ""
	for (section in Sections) {

		# Since V4.80, the [Content Info] is no longer supplied in Linux
		# if (FILEFORMAT == "W" && section == "Content Info")

		if (section == "Content Info")
			continue

		if (Sections[section] == 0)
			missing = missing "\n\t[" section "]"
	}
	if (missing != "") {
		logEND()
		printf("\tERROR: Missing Sections: %s\n", missing)
		status += 1
	}

	# [Component Definitions]
	# Check completeness of keywords defined in ComponentDefinition[]

	missing = ""
	for (def in ComponentDefinition) {
		if (! (def in ComponentDefinition2))
			missing = missing "\n\t'" def "=" ComponentDefinition[def] "'"
	}
	if (missing != "") {
		logEND()
		printf("\tERROR: Missing in [Component Definition]: %s\n", missing)
		status += 1
	}

	# [Physical Definition] and [Functional Definition]
	# Test completeness of keywords
	# Phys: 1 - 4, Func 1 - 12

	missing = ""
	for (i = 1; i <= 4; i++) {
		if (!(i in PhysicalDefinition2))
			missing = sprintf("%s\n\t'Type.%d=%d %s'", missing, i, i, hysicalDefinition[i])
	}
	if (missing != "") {
		logEND()
		printf("\tERROR: Missing in [Physical Definition]: %s\n", missing)
		status += 1
	}

	missing = ""
	for (i = 1; i <= 12; i++) {
		if (!(i in FunctionalDefinition2))
			missing = sprintf("%s\n\t'Type.%d=%d %s'", missing, i, i, FunctionalDefinition[i])
	}
	if (missing != "") {
		logEND()
		printf("\tERROR: Missing in [Functional Definition]: %s\n", missing)
		status += 1
	}

	# [Content Info] and Nr* counters
	# Check completeness of counter keywords appearing in file
	# and compare the counters reported with those actually accumulated.

	# Since V4.80, the [Content Info] is no longer supplied in Linux
	# If there is a [Content Info] then check it

	if (FILEFORMAT == "L" && HaveContentInfo) {
		status += compareTheCounter("CompDefFound")
		status += compareTheCounter("PhysNodeDefFound")
		status += compareTheCounter("PhysFuncDefFound")
		status += compareTheCounter("PhysContDefFound")
		status += compareTheCounter("FuncContDefFound")
		status += compareTheCounter("InfoTableDefFound")
		status += compareTheCounter("InfoFwObjectDefFound")

	} else {
		# Copy these counter to allow subsequent comparison

		ContentInfo[1, "NrPContainer"] = ContentInfo[1, "PhysContDefFound"]
		ContentInfo[1, "NrFContainer"] = ContentInfo[1, "FuncContDefFound"]
		ContentInfo[1, "NrComponents"] = ContentInfo[1, "InfoTableDefFound"]
		ContentInfo[1, "NrUpdObjects"] = ContentInfo[1, "InfoFwObjectDefFound"]

		status += compareTheCounter("NrPContainer")
		status += compareTheCounter("NrFContainer")
		status += compareTheCounter("NrComponents")
		status += compareTheCounter("NrUpdObjects")
	}

	# Print [Physical Container] and [Functional Container] as tables

	pcnt = printChildren(PhysicalContainer, 0)
	fcnt = printChildren(FunctionalContainer, 0)
	if (pcnt != fcnt) {
		logEND()
		printf("\tERROR: Unequal physical/functional components '%d/%d'!\n", pcnt, fcnt)
		status += 1
	}

	exit (status > 0 ? 1 : 0)
}
# EOF
